
#include <linux/wanpipe_includes.h>
#include <linux/wanpipe_defines.h>
#include <linux/wanpipe_debug.h>
#include <linux/wanpipe_common.h>
#include <linux/wanpipe_cfg.h>
#include <linux/proc_fs.h>
#include "wanpipe_lapb.h"

#define PROC_BUFSZ 4000   

#ifdef CONFIG_PROC_FS

extern struct lapb_x25_register_struct x25_protocol;
extern void lapb_link_down (struct net_device *dev);


/* Methods for preparing data for reading proc entries */

#ifdef LINUX_2_6
static int config_get_info(struct seq_file *m, void *v);
static int status_get_info(struct seq_file *m, void *v);
static int lapb_dev_get_info(struct seq_file *m, void *v);
static int lapb_map_get_info(struct seq_file *m, void *v);
#else
static int config_get_info(char* buf, char** start, off_t offs, int len);
static int status_get_info(char* buf, char** start, off_t offs, int len);
static int lapb_dev_get_info(char* buf, char** start, off_t offs, int len);
static int lapb_map_get_info(char* buf, char** start, off_t offs, int len);

/* Proc filesystem interface */
static int lapb_proc_perms(struct inode *, int);
static int lapb_proc_dev_perms (struct inode* inode, int op);

static ssize_t lapb_proc_read(struct file* file, char* buf, size_t count, loff_t *ppos);
static ssize_t lapb_proc_write(struct file* file, const char* buf, size_t count, loff_t *ppos);

#endif
/* Miscellaneous */

/*
 *	Structures for interfacing with the /proc filesystem.
 *	Router creates its own directory /proc/net/router with the folowing
 *	entries:
 *	config		device configuration
 *	status		global device statistics
 *	<device>	entry for each WAN device
 */

/*
 *	Generic /proc/net/router/<file> file and inode operations 
 */

#ifdef LINUX_2_6


static int config_get_info_open(struct inode *inode, struct file *file)
{
	return single_open(file, config_get_info, PDE(inode)->data);
}

static struct file_operations config_get_info_fops = {
	.owner	 = THIS_MODULE,
	.open	 = config_get_info_open,
	.read	 = seq_read,
	.llseek	 = seq_lseek,
	.release = seq_release,
};

static int status_get_info_open(struct inode *inode, struct file *file)
{
	return single_open(file, status_get_info, PDE(inode)->data);
}

static struct file_operations status_get_info_fops = {
	.owner	 = THIS_MODULE,
	.open	 = status_get_info_open,
	.read	 = seq_read,
	.llseek	 = seq_lseek,
	.release = seq_release,
};



static int lapb_map_get_open(struct inode *inode, struct file *file)
{
	return single_open(file, lapb_map_get_info, PDE(inode)->data);
}

static struct file_operations lapb_map_get_fops = {
	.owner	 = THIS_MODULE,
	.open	 = lapb_map_get_open,
	.read	 = seq_read,
	.llseek	 = seq_lseek,
	.release = seq_release,
};

static int lapb_dev_get_open(struct inode *inode, struct file *file)
{
	return single_open(file, lapb_dev_get_info, PDE(inode)->data);
}

static struct file_operations lapb_dev_get_fops = {
	.owner	 = THIS_MODULE,
	.open	 = lapb_dev_get_open,
	.read	 = seq_read,
	.llseek	 = seq_lseek,
	.release = seq_release,
};



#else
static struct file_operations lapb_fops =
{
	read:		lapb_proc_read,
};

static struct inode_operations lapb_inode =
{
	permission:	lapb_proc_perms,
};

/*
 *	/proc/net/router/<device> file operations
 */

static struct file_operations lapb_dev_fops =
{
	read:		lapb_proc_read,
	write:		lapb_proc_write,
	//ioctl:		wanrouter_ioctl,
};

static struct inode_operations lapb_dev_inode =
{
	permission:	lapb_proc_dev_perms,
};

#endif

/*
 *	/proc/net/wanrouter/x25 
 */

static struct proc_dir_entry *proc_lapb;
extern struct proc_dir_entry *proc_router;

/* Strings */
#define PROC_CONF_TEMP "| %-14s| %-14s| %-7s|%-5d|%-8d|%-5d|%-4d|%-4d|%-4d|\n"
static char conf_hdr[] =
        "------------------------------------------------------------------------------\n"
	"| Device name   |    mstr dev   | usedby | mode| window | mtu | T1 | T2 | N2 |\n"
	"------------------------------------------------------------------------------\n";

static char end_hdr[] = "------------------------------------------------------------------------------\n";


#define PROC_STAT_TEMP "| %-14s| %-14s| %-7s|         %-15s           |\n"
static char stat_hdr[] =
        "------------------------------------------------------------------------------\n"
	"| Device name   |    mstr dev   | usedby |              State                |\n"
	"------------------------------------------------------------------------------\n";

//static char stat_hdr[] =
//	"Device name    |protocol|station|interface|clocking|baud rate"
//	"| MTU |ndev|link state\n";


/*
 *	Interface functions
 */

/*
 *	Initialize router proc interface.
 */
#define LAPB_DIR_NAME "lapb"

int lapb_proc_init (void)
{
	struct proc_dir_entry *p;

	if (!proc_router)
		return -ENODEV;
	
	proc_lapb = proc_mkdir(LAPB_DIR_NAME, proc_router);
	if (!proc_lapb)
		goto fail;

	p = create_proc_entry("config",0,proc_lapb);
	if (!p)
		goto fail_config;
	
#ifdef LINUX_2_6
	p->proc_fops = &config_get_info_fops;
#else
	p->proc_fops = &lapb_fops;
	p->proc_iops = &lapb_inode;
	p->get_info = config_get_info;
#endif
	p = create_proc_entry("status",0,proc_lapb);
	if (!p)
		goto fail_stat;

#ifdef LINUX_2_6
	p->proc_fops = &status_get_info_fops;
#else
	p->proc_fops = &lapb_fops;
	p->proc_iops = &lapb_inode;
	p->get_info = status_get_info;
#endif
	
	return 0;
	
fail_stat:
	remove_proc_entry("config", proc_lapb);
fail_config:
	remove_proc_entry(LAPB_DIR_NAME, proc_router);
fail:
	return -ENOMEM;
}

/*
 *	Clean up router proc interface.
 */

void lapb_proc_cleanup (void)
{
	remove_proc_entry("config", proc_lapb);
	remove_proc_entry("status", proc_lapb);
	remove_proc_entry(LAPB_DIR_NAME,proc_router);
}

int lapb_proc_check_use(lapb_cb* lapb)
{

	if (proc_lapb){
		if (atomic_read(&proc_lapb->count)){
			return 1;
		}
	}

	return 0;
}

/*
 *	Add directory entry for WAN device.
 */

int lapb_proc_add (lapb_cb* lapb)
{
	char tmp_name[100];
	sprintf(tmp_name,"%s_map",lapb->dev->name);

	lapb->dent = create_proc_entry(lapb->dev->name, 0666, proc_lapb);
	if (!lapb->dent)
		return -ENOMEM;

#ifdef LINUX_2_6
	lapb->dent->proc_fops	= &lapb_dev_get_fops;
#else
	lapb->dent->proc_fops	= &lapb_dev_fops;
	lapb->dent->proc_iops	= &lapb_dev_inode;
	lapb->dent->get_info	= lapb_dev_get_info;
#endif
	lapb->dent->data	= lapb;

	
	lapb->dent = create_proc_entry(tmp_name, 0, proc_lapb);
	if (!lapb->dent){
		remove_proc_entry(lapb->dev->name, proc_lapb);
		return -ENOMEM;
	}

#ifdef LINUX_2_6
	lapb->dent->proc_fops	= &lapb_map_get_fops;
#else
	lapb->dent->proc_fops	= &lapb_fops;
	lapb->dent->proc_iops	= &lapb_inode;
	lapb->dent->get_info	= lapb_map_get_info;
#endif
	lapb->dent->data	= lapb;

	return 0;
}

/*
 *	Delete directory entry for WAN device.
 */
 
int lapb_proc_delete(lapb_cb* lapb)
{
	char tmp_name[100];
	sprintf(tmp_name,"%s_map",lapb->dev->name);
	remove_proc_entry(lapb->dev->name, proc_lapb);
	remove_proc_entry(tmp_name, proc_lapb);
	lapb->dent=NULL;
	return 0;
}

/****** Proc filesystem entry points ****************************************/

/*
 *	Verify access rights.
 */
#ifndef LINUX_2_6
static int lapb_proc_perms (struct inode* inode, int op)
{
	return 0;
}

static int lapb_proc_dev_perms (struct inode* inode, int op)
{
	int mode = inode->i_mode;
	if (!current->euid){
		mode >>= 6;
		
	}else if (in_egroup_p(0)){
		mode >>= 3;
	}
	
	if ((mode & op & 0007) == op)
		return 0;

	return -EACCES;
}




/*
 *	Read router proc directory entry.
 *	This is universal routine for reading all entries in /proc/net/wanrouter
 *	directory.  Each directory entry contains a pointer to the 'method' for
 *	preparing data for that entry.
 *	o verify arguments
 *	o allocate kernel buffer
 *	o call get_info() to prepare data
 *	o copy data to user space
 *	o release kernel buffer
 *
 *	Return:	number of bytes copied to user space (0, if no data)
 *		<0	error
 */

static ssize_t lapb_proc_read(struct file* file, char* buf, size_t count,
				loff_t *ppos)
{
	struct inode *inode = file->f_dentry->d_inode;
	struct proc_dir_entry* dent;
	char* page;
	int len;

	if (count <= 0)
		return 0;
		
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
	 dent = inode->i_private;
#else
	 dent = inode->u.generic_ip;
#endif
			
	
	if ((dent == NULL) || (dent->get_info == NULL))
		return 0;
		
	page = kmalloc(count, GFP_KERNEL);
	if (page == NULL)
		return -ENOBUFS;
		
	len = dent->get_info(page, dent->data, file->f_pos, count);
	if (len) {
		if(copy_to_user(buf, page, len)){
			kfree(page);
			return -EFAULT;
		}
		file->f_pos += len;
	}
	kfree(page);
	return len;
}

static ssize_t lapb_proc_write(struct file* file, const char* buf, size_t count,
				loff_t *ppos)
{
	struct inode *inode = file->f_dentry->d_inode;
	struct proc_dir_entry* dent;
	char* page;
	lapb_cb* lapb;

	if (count <= 0 || count > 100)
		return 0;
		
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
	 dent = inode->i_private;
#else
	 dent = inode->u.generic_ip;
#endif
			
	if ((dent == NULL) || (dent->get_info == NULL))
		return 0;
		
	if ((lapb=dent->data) == NULL)
		return 0;
	
	page = kmalloc(count, GFP_KERNEL);
	if (page == NULL)
		return -ENOBUFS;

	if (copy_from_user(page, buf, count)){
		kfree(page);
		return -EINVAL;
	}

	page[count-1] = '\0';

	if (strcmp(page,"flush") == 0){
		printk (KERN_INFO "%s: Flushing proc statistics.\n",lapb->dev->name);
		memset(&lapb->lapbstats,0,sizeof(struct lapb_protocol_stats));
		memset(&lapb->ifstats,0,sizeof(struct net_device_stats));
	
	}
	
	kfree(page);
	return count;
}
#endif

/*
 *	Prepare data for reading 'Config' entry.
 *	Return length of data.
 */
extern struct net_device *lapb_list;

#ifdef LINUX_2_6
static int config_get_info(struct seq_file *m, void *v)
#else
static int config_get_info(char* buf, char** start, off_t offs, int len)
#endif
{
	lapb_cb* lapb;
	struct net_device *tmp_dev;
	unsigned int kmem,skb_mem;
	PROC_ADD_DECL(m);
	
	PROC_ADD_INIT(m,buf,offs,len);
	PROC_ADD_LINE(m,conf_hdr);

	for (tmp_dev = lapb_list; tmp_dev;
 	     tmp_dev= *(struct net_device **)tmp_dev->priv){

		lapb = tmp_dev->priv;
		if (!lapb)
			continue;

		PROC_ADD_LINE(m,
		         PROC_CONF_TEMP,
			 lapb->dev->name,
			 lapb->tx_dev->name,
			 lapb->x25_link ? "X25" : lapb->sk_id ? "API" : "TCP/IP",
			 lapb->mode,
			 lapb->window,
			 lapb->mtu,
			 lapb->t1/HZ,
			 lapb->t2/HZ,
			 lapb->n2);
	}

	PROC_ADD_LINE(m,end_hdr);

	kmem=atomic_read(&lapb_mem_alloc);
	skb_mem=atomic_read(&lapb_skb_alloc);
	
	PROC_ADD_LINE(m,
		      "\nMemory Used=%d : Skb Used=%d : Total=%d\n",
		      kmem,skb_mem,kmem+skb_mem);

	
	PROC_ADD_RET(m);
}

/*
 *	Prepare data for reading 'Status' entry.
 *	Return length of data.
 */
#ifdef LINUX_2_6
static int status_get_info(struct seq_file *m, void *v)
#else
static int status_get_info(char* buf, char** start, off_t offs, int len)
#endif
{
	lapb_cb* lapb;
	struct net_device *tmp_dev;
	PROC_ADD_DECL(m);

	PROC_ADD_INIT(m,buf,offs,len);
	PROC_ADD_LINE(m,stat_hdr);
	for (tmp_dev = lapb_list; tmp_dev;
 	     tmp_dev= *(struct net_device **)tmp_dev->priv){

		lapb = tmp_dev->priv;
		if (!lapb)
			continue;

		PROC_ADD_LINE(m,
			 PROC_STAT_TEMP,
			 lapb->dev->name,
			 lapb->tx_dev->name,
			 lapb->x25_link ? "X25" : lapb->sk_id ? "API" : "TCP/IP",
			 lapb->state == LAPB_STATE_DISCONNECTED ? "Disconnected" : 
			 lapb->state == LAPB_STATE_CONNECTING ? "Connecting" :
			 lapb->state == LAPB_STATE_DISCONNECTING ? "Disconnecting" : 
			 lapb->state == LAPB_STATE_CONNECTED ? "Connected" : "Frame Reject");
	}
	
	
	PROC_ADD_LINE(m, end_hdr);
	PROC_ADD_RET(m);

}

/*
 *	Prepare data for reading <device> entry.
 *	Return length of data.
 *
 *	On entry, the 'start' argument will contain a pointer to WAN device
 *	data space.
 */

#define PROC_DEV_FORMAT "| %-12s|  %-17d|   %-17d|\n"
#define PROC_DEV_FORMAT1 "| %-15s : %-10d | %-15s : %-10d |\n"
unsigned char dev_line[]= "--------------------------------------------------------\n";
unsigned char dev_line1[]= "---------------------------------------------------------------\n";
unsigned char dev_hdr[] = "| Packet Type |     Received      |    Transmitted     |\n";

#ifdef LINUX_2_6
static int lapb_dev_get_info(struct seq_file *m, void *v)
#else
static int lapb_dev_get_info(char* buf, char** start, off_t offs, int len)
#endif
{
	lapb_cb* lapb = PROC_GET_DATA();
	PROC_ADD_DECL(m);

	if ((lapb == NULL))
		return 0;

	PROC_ADD_INIT(m,buf,offs,len);

	//PROC_ADD_LINE(m,mdev_line);
	PROC_ADD_LINE(m,dev_hdr);
	PROC_ADD_LINE(m,dev_line);
	
	PROC_ADD_LINE(m, 
		 PROC_DEV_FORMAT,
		"I-Frame", lapb->lapbstats.ifrm_rx,lapb->lapbstats.ifrm_tx);
	
	PROC_ADD_LINE(m,PROC_DEV_FORMAT,
		"RR", lapb->lapbstats.rr_rx,lapb->lapbstats.rr_tx);
	PROC_ADD_LINE(m, PROC_DEV_FORMAT,
		"RNR", lapb->lapbstats.rnr_rx,lapb->lapbstats.rnr_tx);
	PROC_ADD_LINE(m, PROC_DEV_FORMAT,
		"REJ", lapb->lapbstats.rej_rx,lapb->lapbstats.rej_tx);

	PROC_ADD_LINE(m, PROC_DEV_FORMAT,
		"SABM", lapb->lapbstats.sabm_rx,lapb->lapbstats.sabm_tx);		

	PROC_ADD_LINE(m,PROC_DEV_FORMAT,
		"DISC", lapb->lapbstats.disc_rx,lapb->lapbstats.disc_tx);
	
	PROC_ADD_LINE(m,PROC_DEV_FORMAT,
		"UA", lapb->lapbstats.ua_rx,lapb->lapbstats.ua_tx);

	PROC_ADD_LINE(m,PROC_DEV_FORMAT,
		"DM", lapb->lapbstats.dm_rx,lapb->lapbstats.dm_tx);

	PROC_ADD_LINE(m,PROC_DEV_FORMAT,
		"FRMR", lapb->lapbstats.frmr_rx,lapb->lapbstats.frmr_tx);

	PROC_ADD_LINE(m, PROC_DEV_FORMAT,
		"Illegal", lapb->lapbstats.ill_rx,lapb->lapbstats.ill_tx);

	PROC_ADD_LINE(m,dev_line);
	
	PROC_ADD_LINE(m, PROC_DEV_FORMAT,
		"Total Pkts", (u32)lapb->ifstats.rx_packets,(u32)lapb->ifstats.tx_packets);
	
	PROC_ADD_LINE(m, PROC_DEV_FORMAT,
		"Total Bytes", (u32)lapb->ifstats.rx_bytes,(u32)lapb->ifstats.tx_bytes);

	PROC_ADD_LINE(m, dev_line);

	PROC_ADD_LINE(m, dev_line1);

	PROC_ADD_LINE(m,
			"|      LAPB Protocol Driver Stats                             |\n");

	PROC_ADD_LINE(m, dev_line1);

	PROC_ADD_LINE(m, 
		 "| %-15s : %-10s | %-15s : %-10s |\n", 
		 "Link State", lapb->state == LAPB_STATE_DISCONNECTED ? "Disc" : 
			 	lapb->state == LAPB_STATE_CONNECTING ? "Con.g" :
			   	lapb->state == LAPB_STATE_DISCONNECTING ? "Dis.g" : 
			     	lapb->state == LAPB_STATE_CONNECTED ? "Con.d" : "FR.t", 
		"Dev State", netif_queue_stopped(lapb->dev) ? "Stp" : "Run");
	
	PROC_ADD_LINE(m,PROC_DEV_FORMAT1, 
		 "Write Queue", skb_queue_len(&lapb->write_queue),
		 "Ack Queue", skb_queue_len(&lapb->ack_queue));
	
	PROC_ADD_LINE(m,PROC_DEV_FORMAT1, 
		 "Async Queue", skb_queue_len(&lapb->async_queue),
		 "Tx Kick Cnt", lapb->kick_cnt);

	PROC_ADD_LINE(m, 
		 "| %-15s : 0x%-8lx | %-15s : 0x%-8x |\n", 
		 "Lapb Criticals", lapb->critical,
		 "Lapb Condition", lapb->condition);
	
	PROC_ADD_LINE(m, 
		 "| %-15s : %-10ld | %-15s : %-10d |\n", 
		 "Lapb Awt Kick", (lapb->critical & LAPB_AWAITING_KICK),
		 "N/A", 0);

	PROC_ADD_LINE(m, dev_line1);

	PROC_ADD_LINE(m, 
		 "| %-10s : %-5d | %-10s : %-5d | %-10s : %-4d |\n", 
		 "Lapb Vs", lapb->vs,
		 "Lapb Vr", lapb->vr,
		 "Lapb Va", lapb->va);

	PROC_ADD_LINE(m, dev_line1);

	PROC_ADD_RET(m);
}

#ifdef LINUX_2_6
static int lapb_map_get_info(struct seq_file *m, void *v)
#else
static int lapb_map_get_info(char* buf, char** start, off_t offs, int len)
#endif
{
	lapb_cb* lapb = PROC_GET_DATA();
	char	line[200];
	
	PROC_ADD_DECL(m);

	PROC_ADD_INIT(m,buf,offs,len);

	if ((lapb == NULL))
		return 0;

	PROC_ADD_LINE(m, ":%s", lapb->hw_name);
	
	PROC_ADD_LINE(m, ":%s:%c", lapb->tx_dev->name,
		 lapb->state == LAPB_STATE_DISCONNECTED ? ' ' : '*');
 
	PROC_ADD_LINE(m, ":%s:%c", lapb->dev->name,
		 test_bit(LAPB_DOWN,&lapb->critical) ? 'd' :
		 lapb->state == LAPB_STATE_CONNECTED ? '*' : ' ');

#ifndef LINUX_2_6
	if (lapb->x25_link){
		if (IS_FUNC_CALL(x25_protocol,x25_get_map)){
			x25_protocol.x25_get_map(lapb->x25_link,
							m);
		}
	}else{
		PROC_ADD_LINE(m, "%15s%s\n", 
		 	lapb->label, line);
	}
#else
	PROC_ADD_LINE(m, "%15s%s\n", 
		 	lapb->label, line);
#endif

	PROC_ADD_RET(m);
}



#endif
