#include <linux/wanpipe_includes.h>
#include <linux/wanpipe_defines.h>
#include <linux/wanpipe_debug.h>
#include <linux/wanpipe_common.h>
#include <linux/wanpipe_cfg.h>
#include <linux/proc_fs.h>
#include "wanpipe_x25.h"

#define PROC_BUFSZ 4000   

#ifdef CONFIG_PROC_FS


#ifdef LINUX_2_6
/* Methods for preparing data for reading proc entries */
static int config_get_info(struct seq_file *m, void *v);
static int status_get_info(struct seq_file *m, void *v);

static int x25_link_config_get_info(struct seq_file *m, void *v);
static int x25_link_status_get_info(struct seq_file *m, void *v);
static int x25_link_events_get_info(struct seq_file *m, void *v);

static int x25_dev_get_info(struct seq_file *m, void *v);

#else
/* Methods for preparing data for reading proc entries */
static int config_get_info(char* buf, char** start, off_t offs, int len);
static int status_get_info(char* buf, char** start, off_t offs, int len);

static int x25_link_config_get_info(char* buf, char** start, off_t offs, int len);
static int x25_link_status_get_info(char* buf, char** start, off_t offs, int len);
static int x25_link_events_get_info(char* buf, char** start, off_t offs, int len);

static int x25_dev_get_info(char* buf, char** start, off_t offs, int len);

/* Proc filesystem interface */
static int x25_proc_perms(struct inode *, int);
static int x25_proc_dev_perms (struct inode* inode, int op);

static ssize_t x25_proc_read(struct file* file, char* buf, size_t count, loff_t *ppos);

static ssize_t x25_proc_link_write(struct file* file, const char* buf, size_t count, loff_t *ppos);
static ssize_t x25_proc_lcn_write(struct file* file, const char* buf, size_t count, loff_t *ppos);

#endif

/* Miscellaneous */

/*
 *	Structures for interfacing with the /proc filesystem.
 *	Router creates its own directory /proc/net/router with the folowing
 *	entries:
 *	config		device configuration
 *	status		global device statistics
 *	<device>	entry for each WAN device
 */

/*
 *	Generic /proc/net/wanrouter/x25/<file> file and inode operations 
 */


#ifdef LINUX_2_6

static int config_get_info_open(struct inode *inode, struct file *file)
{
	return single_open(file, config_get_info, PDE(inode)->data);
}

static struct file_operations config_get_info_fops = {
	.owner	 = THIS_MODULE,
	.open	 = config_get_info_open,
	.read	 = seq_read,
	.llseek	 = seq_lseek,
	.release = seq_release,
};

static int status_get_info_open(struct inode *inode, struct file *file)
{
	return single_open(file, status_get_info, PDE(inode)->data);
}

static struct file_operations status_get_info_fops = {
	.owner	 = THIS_MODULE,
	.open	 = status_get_info_open,
	.read	 = seq_read,
	.llseek	 = seq_lseek,
	.release = seq_release,
};

static int x25_link_config_get_info_open(struct inode *inode, struct file *file)
{
	return single_open(file, x25_link_config_get_info, PDE(inode)->data);
}

static struct file_operations x25_link_config_get_info_fops = {
	.owner	 = THIS_MODULE,
	.open	 = x25_link_config_get_info_open,
	.read	 = seq_read,
	.llseek	 = seq_lseek,
	.release = seq_release,
};

static int x25_link_status_get_info_open(struct inode *inode, struct file *file)
{
	return single_open(file, x25_link_status_get_info, PDE(inode)->data);
}

static struct file_operations x25_link_status_get_info_fops = {
	.owner	 = THIS_MODULE,
	.open	 = x25_link_status_get_info_open,
	.read	 = seq_read,
	.llseek	 = seq_lseek,
	.release = seq_release,
};

static int x25_link_events_get_info_open(struct inode *inode, struct file *file)
{
	return single_open(file, x25_link_events_get_info, PDE(inode)->data);
}

static struct file_operations x25_link_events_get_info_fops = {
	.owner	 = THIS_MODULE,
	.open	 = x25_link_events_get_info_open,
	.read	 = seq_read,
	.llseek	 = seq_lseek,
	.release = seq_release,
};

static int x25_dev_get_info_open(struct inode *inode, struct file *file)
{
	return single_open(file, x25_dev_get_info, PDE(inode)->data);
}

static struct file_operations x25_dev_get_info_fops = {
	.owner	 = THIS_MODULE,
	.open	 = x25_dev_get_info_open,
	.read	 = seq_read,
	.llseek	 = seq_lseek,
	.release = seq_release,
};

#else
static struct file_operations x25_fops =
{
	read:		x25_proc_read,
};

static struct inode_operations x25_inode =
{
	permission:	x25_proc_perms,
};

/*
 *	/proc/net/wanrouter/x25/<device> file operations
 */

static struct file_operations x25_link_fops =
{
	read:		x25_proc_read,
	write: 		x25_proc_link_write,
};



static struct file_operations x25_dev_fops =
{
	read:		x25_proc_read,
	write: 		x25_proc_lcn_write,
};

static struct inode_operations x25_dev_inode =
{
	permission:	x25_proc_dev_perms,
};

#endif

/*
 *	/proc/net/wanrouter/x25 
 */

static struct proc_dir_entry *proc_x25;
extern struct proc_dir_entry *proc_router;
extern  x25_link_t *x25_link_list;


/* Strings */

static char end_hdr[] = "------------------------------------------------------------------------------\n";


#define PROC_STAT_TEMP "| %-14s| %-14s| %-7s|         %-15s           |\n"
/*
 *	Interface functions
 */

/*
 *	Initialize router proc interface.
 */
#define X25_DIR_NAME "x25"

int x25_proc_init (void)
{
	struct proc_dir_entry *p;

	if (!proc_router)
		return -ENODEV;
	
	proc_x25 = proc_mkdir(X25_DIR_NAME, proc_router);
	if (!proc_x25)
		goto x25_proc_init_fail;

	p = create_proc_entry("config",0,proc_x25);
	if (!p)
		goto x25_proc_init_fail_config;
#ifdef LINUX_2_6	
	p->proc_fops = &config_get_info_fops;
#else
	p->proc_fops = &x25_fops;
	p->proc_iops = &x25_inode;
	p->get_info = config_get_info;
#endif
	p = create_proc_entry("status",0,proc_x25);
	if (!p)
		goto x25_proc_init_fail_stat;
		
#ifdef LINUX_2_6
	p->proc_fops = &status_get_info_fops;
#else
	p->proc_fops = &x25_fops;
	p->proc_iops = &x25_inode;
	p->get_info = status_get_info;
#endif
	return 0;
	
x25_proc_init_fail_stat:
	remove_proc_entry("config", proc_x25);

x25_proc_init_fail_config:
	remove_proc_entry(X25_DIR_NAME, proc_router);

x25_proc_init_fail:
	return -ENOMEM;
}

/*
 *	Clean up router proc interface.
 */

void x25_proc_cleanup (void)
{
	remove_proc_entry("config", proc_x25);
	remove_proc_entry("status", proc_x25);
	remove_proc_entry("events", proc_x25);
	remove_proc_entry(X25_DIR_NAME,proc_router);
}



int x25_proc_link_init (x25_link_t *x25_link)
{
	struct proc_dir_entry *p;

	if (!proc_router)
		return -ENODEV;

	x25_link->proc_dir = proc_mkdir(x25_link->name, proc_x25);
	if (!x25_link->proc_dir)
		goto x25_link_init_fail;

	p = create_proc_entry("config",0,x25_link->proc_dir);
	if (!p)
		goto x25_link_init_fail_config;
		
#ifdef LINUX_2_6
	p->proc_fops = &x25_link_config_get_info_fops;
#else
	p->proc_fops = &x25_fops;
	p->proc_iops = &x25_inode;
	p->get_info = x25_link_config_get_info;
#endif
	p->data	= x25_link;
	p = create_proc_entry("status",0666,x25_link->proc_dir);
	if (!p)
		goto x25_link_init_fail_stat;
	
#ifdef LINUX_2_6
	p->proc_fops = &x25_link_status_get_info_fops;
#else
	p->proc_fops = &x25_link_fops;
	p->proc_iops = &x25_dev_inode;
	p->get_info = x25_link_status_get_info;
#endif
	p->data	= x25_link;

	p = create_proc_entry("events",0666,x25_link->proc_dir);
	if (!p)
		goto x25_link_init_fail_events;
		
#ifdef LINUX_2_6
	p->proc_fops = &x25_link_events_get_info_fops;
#else
	p->proc_fops = &x25_link_fops;
	p->proc_iops = &x25_dev_inode;
	p->get_info = x25_link_events_get_info;
#endif
	p->data	= x25_link;
	
	return 0;

x25_link_init_fail_events:
	remove_proc_entry("status", x25_link->proc_dir);
	
x25_link_init_fail_stat:
	remove_proc_entry("config", x25_link->proc_dir);

x25_link_init_fail_config:
	remove_proc_entry(x25_link->name, proc_x25);

x25_link_init_fail:
	printk(KERN_INFO "(Debug): ERROR !!!!!! X25 link init fail\n");
	return -ENOMEM;

}


int x25_proc_link_cleanup(x25_link_t *x25_link)
{

	if (!x25_link->proc_dir || !proc_x25){
		DEBUG_EVENT("%s: CRITICAL ERROR: Not proc dir!\n",
				__FUNCTION__);
		return 0;
	}

	if (x25_link->proc_dir){
		remove_proc_entry("config", x25_link->proc_dir);
	}

	if (x25_link->proc_dir){	
		remove_proc_entry("status", x25_link->proc_dir);
	}

	if (x25_link->proc_dir){	
		remove_proc_entry("events", x25_link->proc_dir);
	}

	if (proc_x25){
		remove_proc_entry(x25_link->name,proc_x25);
	}
	return 0;
}

/*
 *	Add directory entry for WAN device.
 */

int x25_proc_add (x25_lcn_t* x25_lcn, struct proc_dir_entry *proc_dir)
{
	x25_lcn->dent = create_proc_entry(x25_lcn->dev->name, 0666, proc_dir);
	if (!x25_lcn->dent)
		return -ENOMEM;
#ifdef LINUX_2_6
	x25_lcn->dent->proc_fops = &x25_dev_get_info_fops;
#else
	x25_lcn->dent->proc_fops = &x25_dev_fops;
	x25_lcn->dent->proc_iops = &x25_dev_inode;
	x25_lcn->dent->get_info	= x25_dev_get_info;
#endif
	x25_lcn->dent->data	= x25_lcn;
	return 0;
}

int x25_proc_usage_check(void)
{
	if (proc_x25){
		return atomic_read(&proc_x25->count);
	}

	return 0;
}

/*
 *	Delete directory entry for WAN device.
 */
 
int x25_proc_delete(x25_lcn_t* x25_lcn,struct proc_dir_entry *proc_dir)
{	
	remove_proc_entry(x25_lcn->dev->name, proc_dir);
	return 0;
}

/****** Proc filesystem entry points ****************************************/

#ifndef LINUX_2_6
/*
 *	Verify access rights.
 */

static int x25_proc_perms (struct inode* inode, int op)
{
	return 0;
}

static int x25_proc_dev_perms (struct inode* inode, int op)
{

	int mode = inode->i_mode;
	if (!current->euid){
		mode >>= 6;
		
	}else if (in_egroup_p(0)){
		mode >>= 3;
	}
	
	if ((mode & op & 0007) == op)
		return 0;

	return -EACCES;
}


/*
 *	Read router proc directory entry.
 *	This is universal routine for reading all entries in /proc/net/wanrouter
 *	directory.  Each directory entry contains a pointer to the 'method' for
 *	preparing data for that entry.
 *	o verify arguments
 *	o allocate kernel buffer
 *	o call get_info() to prepare data
 *	o copy data to user space
 *	o release kernel buffer
 *
 *	Return:	number of bytes copied to user space (0, if no data)
 *		<0	error
 */

static ssize_t x25_proc_read(struct file* file, char* buf, size_t count,
				loff_t *ppos)
{
	struct inode *inode = file->f_dentry->d_inode;
	struct proc_dir_entry* dent;
	char* 		page = NULL;
	int 		len;

	if (count <= 0)
		return 0;
		
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
	        dent = inode->i_private;
#else
		dent = inode->u.generic_ip;
#endif

	if ((dent == NULL) || (dent->get_info == NULL))
		return 0;
		
	page=kmalloc(count, GFP_KERNEL);
	if (page == NULL)
		return -ENOBUFS;
		
	len = dent->get_info(page, dent->data, file->f_pos, count);
	if (len) {
		if(copy_to_user(buf, page, len)){
			kfree(page);
			return -EFAULT;
		}
		file->f_pos += len;
	}
	kfree(page);
	return len;
}



/*
 *	Prepare data for reading 'Config' entry.
 *	Return length of data.
 */


static ssize_t x25_proc_link_write(struct file* file, const char* buf, size_t count, loff_t *ppos)
{
	struct inode *inode = file->f_dentry->d_inode;
	struct proc_dir_entry* dent;
	char* 		page = NULL;
	x25_link_t 	*x25_link;

	if (count <= 0 || count >= 100 )
		return 0;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
        dent = inode->i_private;
#else
        dent = inode->u.generic_ip;
#endif

	if ((dent == NULL) || (dent->get_info == NULL))
		return 0;
	
	if ((x25_link = dent->data) == NULL){
		return 0;
	}
	
	page=kmalloc(count, GFP_KERNEL);
	if (page == NULL)
		return -ENOBUFS;

	if (copy_from_user(page, buf, count)){
		kfree(page);
		return -EINVAL;
	}
	page[count-1] = '\0';
	
	if (strcmp(page,"flush") == 0){
		printk (KERN_INFO "%s: Flushing proc statistics.\n",x25_link->name);
		memset(&x25_link->x25stats,0,sizeof(struct x25_prot_stats));
	}

	kfree(page);
	return count;
}

static ssize_t x25_proc_lcn_write(struct file* file, const char* buf, size_t count, loff_t *ppos)
{
	struct inode *inode = file->f_dentry->d_inode;
	struct proc_dir_entry* dent;
	char* 		page = NULL;
	x25_lcn_t 	*x25_lcn;

	if (count <= 0 || count >= 100 )
		return 0;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
        dent = inode->i_private;
#else
        dent = inode->u.generic_ip;
#endif

	if ((dent == NULL) || (dent->get_info == NULL))
		return 0;
	
	if ((x25_lcn = dent->data) == NULL){
		return 0;
	}
	
	page=kmalloc(count, GFP_KERNEL);
	if (page == NULL)
		return -ENOBUFS;

	if (copy_from_user(page, buf, count)){
		kfree(page);
		return -EINVAL;
	}
	page[count-1] = '\0';
	
	if (strcmp(page,"flush") == 0){
		printk (KERN_INFO "%s: Flushing proc statistics.\n",x25_lcn->dev->name);
		memset(&x25_lcn->ifstats,0,sizeof(struct net_device_stats));
	}

	kfree(page);
	return count;
}
#endif



#define PROC_CONF_TEMP "| %-9s| %-13s| %-6d| %-6d| %-2d| %-4d| %-5d| %-16s|\n"
static char conf_hdr[] =
        "------------------------------------------------------------------------------\n"
	"| X25 Link |  master dev  |no chan|modulus|DTE| MTU |window|     State       |\n"
	"------------------------------------------------------------------------------\n";
#ifdef LINUX_2_6
static int config_get_info(struct seq_file *m, void *v)
#else
static int config_get_info(char* buf, char** start, off_t offs, int len)
#endif
{
	x25_link_t* 	x25_link=NULL;
	unsigned int kmem,skb_mem;

	PROC_ADD_DECL(m)

	PROC_ADD_INIT(m,buf,offs,len);
	PROC_ADD_LINE(m, conf_hdr);

	for (x25_link = x25_link_list; x25_link; x25_link = x25_link->next){ 

		PROC_ADD_LINE(m,
			 PROC_CONF_TEMP,
			 x25_link->name,
			 x25_link->master->name,
			 x25_link->dev_cnt,
			 (x25_link->cfg.mode & GFI_MOD_8_BIT_SET) ? 
						MOD8_MODULUS : MOD128_MODULUS,
			 x25_link->cfg.dte,		
			 x25_link->cfg.default_packet_size,
			 x25_link->cfg.packet_window,
			 x25_link->state == X25_ST_PKT_LVL_RDY_R1 ? "Link Ready" :
				x25_link->state == X25_ST_RESTART_REQ_SNT_R2 ? "Restart Recvd" :
				x25_link->state == X25_ST_RESTART_REQ_SNT_R2 ? "Restart Snt" :
				"Link Not Ready");
	}

	PROC_ADD_LINE(m,end_hdr);

	kmem=atomic_read(&x25_mem_alloc);
	skb_mem=atomic_read(&x25_skb_alloc);
	
	PROC_ADD_LINE(m,
			"\nMemory Used=%d : Skb Used=%d : Total=%d\n",
		  	kmem,skb_mem,kmem+skb_mem);

	PROC_ADD_RET(m);
}

/*
 *	Prepare data for reading 'Status' entry.
 *	Return length of data.
 */

#ifdef LINUX_2_6
static int status_get_info(struct seq_file *m, void *v)
#else
static int status_get_info(char* buf, char** start, off_t offs, int len)
#endif
{

#if 0
	int cnt = sizeof(stat_hdr) - 1;
	struct net_device *tmp_dev;
	
	lapb_cb* lapb;
	strcpy(buf, stat_hdr);

	for (tmp_dev = lapb_list; 
 	     tmp_dev != NULL && (cnt < (PROC_BUFSZ - 120)) ; 
 	     tmp_dev= *(struct net_device **)tmp_dev->priv){

		lapb = tmp_dev->priv;
		if (!lapb)
			continue;

		cnt += sprintf(&buf[cnt], PROC_STAT_TEMP,
				lapb->dev->name,
				lapb->tx_dev->name,
				lapb->x25_link ? "X25" : lapb->sk_id ? "API" : "TCP/IP",
				lapb->state == LAPB_STATE_DISCONNECTED ? "Disconnected" : 
				 lapb->state == LAPB_STATE_CONNECTING ? "Connecting" :
				   lapb->state == LAPB_STATE_DISCONNECTING ? "Disconnecting" : 
				     lapb->state == LAPB_STATE_CONNECTED ? "Connected" : "Frame Reject");
	}

	if ((cnt+sizeof(end_hdr)) < (PROC_BUFSZ)){
		strcpy(&buf[cnt],end_hdr);
		cnt += sizeof(end_hdr)-1;
	}
#endif
	return  0; //cnt;
}


#define PROC_LINK_CONF_FORMAT "| %-17s| %-10d| %-17s| %-10d|\n"
#define PROC_LINK_CONF_X_FORMAT "| %-17s| 0x%-8X| %-17s| %-10d|\n"

static char conf_link_hdr1[] =
        "---------------------------------------------------------------\n"
	"| X25 Link Configuration                                      |\n"
	"---------------------------------------------------------------\n";

static char conf_link_line[] =
	"---------------------------------------------------------------\n";

#ifdef LINUX_2_6
static int x25_link_config_get_info(struct seq_file *m, void *v)
#else
static int x25_link_config_get_info(char* buf, char** start, off_t offs, int len)
#endif
{
	x25_link_t*	x25_link = PROC_GET_DATA();
	PROC_ADD_DECL(m)

	WAN_ASSERT2((x25_link==NULL),0);
	
	PROC_ADD_INIT(m,buf,offs,len);
	PROC_ADD_LINE(m,conf_link_hdr1);

	PROC_ADD_LINE(m,
		 PROC_LINK_CONF_X_FORMAT,
		 "Protocol Opt", x25_link->cfg.X25_protocol_options,
		 "Station", x25_link->cfg.dte);
	
	PROC_ADD_LINE(m,
		 PROC_LINK_CONF_X_FORMAT,
		 "Response Opt", x25_link->cfg.X25_response_options,
		 "Modulus", (x25_link->cfg.mode & GFI_MOD_8_BIT_SET) ? 
					MOD8_MODULUS : MOD128_MODULUS);

	PROC_ADD_LINE(m,
		 PROC_LINK_CONF_X_FORMAT,
		 "API Options", x25_link->cfg.X25_API_options,
		 "Lowest SVC", x25_link->cfg.lowest_two_way_channel);

	PROC_ADD_LINE(m,
		 PROC_LINK_CONF_X_FORMAT,
		 "Statistics Opt", x25_link->cfg.X25_statistics_options,
		 "Highest SVC", x25_link->cfg.highest_two_way_channel);

	PROC_ADD_LINE(m,
		 PROC_LINK_CONF_X_FORMAT,
		 "Gen Facility 1", x25_link->cfg.genl_facilities_supported_1,
		 "Pkt Window", x25_link->cfg.packet_window);

	PROC_ADD_LINE(m,
		 PROC_LINK_CONF_X_FORMAT,
		 "Gen Facility 2", x25_link->cfg.genl_facilities_supported_2,
		 "CCITT Compat", x25_link->cfg.CCITT_compatibility);
	
	PROC_ADD_LINE(m,
		 PROC_LINK_CONF_X_FORMAT,
		 "CCITT Facility", x25_link->cfg.CCITT_facilities_supported,
		 "Default Pkt Sz", x25_link->cfg.default_packet_size);

	PROC_ADD_LINE(m,
		 PROC_LINK_CONF_FORMAT,
		 "Non Facility", x25_link->cfg.non_X25_facilities_supported,
		 "Max Pkt Sz", x25_link->cfg.maximum_packet_size);

	PROC_ADD_LINE(m,
		 PROC_LINK_CONF_FORMAT,
		 "Restart T10_T20", x25_link->cfg.T10_T20/HZ,
		 "Call T11_T21", x25_link->cfg.T11_T21/HZ);

	PROC_ADD_LINE(m,
		 PROC_LINK_CONF_FORMAT,
		 "Restart R10_R20", x25_link->cfg.R10_R20, "N/A", 0);

	PROC_ADD_LINE(m,
		 PROC_LINK_CONF_FORMAT,
		 "Reset T12_T22", x25_link->cfg.T12_T22/HZ,
		 "Clear T13_T23", x25_link->cfg.T13_T23/HZ);
	
	PROC_ADD_LINE(m,
		 PROC_LINK_CONF_FORMAT,
		 "Reset R12_R22", x25_link->cfg.R12_R22,
		 "Clear R13_R23", x25_link->cfg.R13_R23);
	
	PROC_ADD_LINE(m,conf_link_line);

	PROC_ADD_LINE(m, "| %-17s| %-41s|\n",
		 "Master Tx Device", x25_link->master->name);
	
	PROC_ADD_LINE(m, conf_link_line);
	
	PROC_ADD_RET(m);
}

#define PROC_LINK_STATS_FORMAT "| %-15s|  %-17d|   %-17d|\n"
unsigned char link_line[]= "-----------------------------------------------------------\n";
unsigned char link_hdr[] = "| Packet Type    |     Received      |    Transmitted     |\n";
	
#ifdef LINUX_2_6
static int x25_link_status_get_info(struct seq_file *m, void *v)
#else
static int x25_link_status_get_info(char* buf, char** start, off_t offs, int len)
#endif
{
	x25_link_t *x25_link = PROC_GET_DATA();
	PROC_ADD_DECL(m)

	if (!x25_link)
		return 0;

	PROC_ADD_INIT(m,buf,offs,len);
	PROC_ADD_LINE(m, link_line);
	PROC_ADD_LINE(m, link_hdr);
	PROC_ADD_LINE(m, link_line);

	PROC_ADD_LINE(m, PROC_LINK_STATS_FORMAT,
		 "Data Pkt", x25_link->x25stats.rx_data_pkt, x25_link->x25stats.tx_data_pkt);
	PROC_ADD_LINE(m, PROC_LINK_STATS_FORMAT,
		 "RR Pkt", x25_link->x25stats.rx_rr, x25_link->x25stats.tx_rr);
	PROC_ADD_LINE(m, PROC_LINK_STATS_FORMAT,
		 "RNR Pkt", x25_link->x25stats.rx_rnr, x25_link->x25stats.tx_rnr);
	PROC_ADD_LINE(m, PROC_LINK_STATS_FORMAT,
		 "REJ Pkt", x25_link->x25stats.rx_rej, x25_link->x25stats.tx_rej);
	PROC_ADD_LINE(m, PROC_LINK_STATS_FORMAT,
		 "REJ Pkt", x25_link->x25stats.rx_rej,x25_link->x25stats.tx_rej);
	PROC_ADD_LINE(m, PROC_LINK_STATS_FORMAT,
		 "Call Req", x25_link->x25stats.rx_call_req,x25_link->x25stats.tx_call_req);
	PROC_ADD_LINE(m, PROC_LINK_STATS_FORMAT,
		 "Call Accept", x25_link->x25stats.rx_call_acpt,x25_link->x25stats.tx_call_acpt);
	PROC_ADD_LINE(m, PROC_LINK_STATS_FORMAT,
		 "Clear Req", x25_link->x25stats.rx_clear_req,x25_link->x25stats.tx_clear_req);
	PROC_ADD_LINE(m, PROC_LINK_STATS_FORMAT,
		 "Clear Conf", x25_link->x25stats.rx_clear_conf,x25_link->x25stats.tx_clear_conf);
	PROC_ADD_LINE(m, PROC_LINK_STATS_FORMAT,
		 "Reset Req", x25_link->x25stats.rx_reset_req,x25_link->x25stats.tx_reset_req);
	PROC_ADD_LINE(m, PROC_LINK_STATS_FORMAT,
		 "Reset Conf", x25_link->x25stats.rx_reset_conf,x25_link->x25stats.tx_reset_conf);
	PROC_ADD_LINE(m, PROC_LINK_STATS_FORMAT,
		 "Restart Req", x25_link->x25stats.rx_restrt_req,x25_link->x25stats.tx_restrt_req);
	PROC_ADD_LINE(m, PROC_LINK_STATS_FORMAT,
		 "Restart Conf", x25_link->x25stats.rx_restrt_conf,x25_link->x25stats.tx_restrt_conf);
	PROC_ADD_LINE(m, PROC_LINK_STATS_FORMAT,
		 "Interrupt Req", x25_link->x25stats.rx_intr_req,x25_link->x25stats.tx_intr_req);
	PROC_ADD_LINE(m, PROC_LINK_STATS_FORMAT,
		 "Interrupt Conf", x25_link->x25stats.rx_intr_conf,x25_link->x25stats.tx_intr_conf);
	PROC_ADD_LINE(m, PROC_LINK_STATS_FORMAT,
		 "Registr Req", x25_link->x25stats.rx_reg_req,x25_link->x25stats.tx_reg_req);
	PROC_ADD_LINE(m, PROC_LINK_STATS_FORMAT,
		 "Registr Conf", x25_link->x25stats.rx_reg_conf,x25_link->x25stats.tx_reg_conf);
	PROC_ADD_LINE(m, PROC_LINK_STATS_FORMAT,
		 "Diagnostic Pkt", x25_link->x25stats.rx_diag,x25_link->x25stats.tx_diag);
	PROC_ADD_LINE(m, PROC_LINK_STATS_FORMAT,
		 "Illegal Pkt", x25_link->x25stats.rx_ill,x25_link->x25stats.tx_ill);
	
	PROC_ADD_LINE(m,link_line);
	PROC_ADD_LINE(m,PROC_LINK_STATS_FORMAT,
		 "Total Pkts", x25_link->x25stats.rx_total,x25_link->x25stats.tx_total);
	PROC_ADD_LINE(m,PROC_LINK_STATS_FORMAT,
		 "Async/Rx Q", skb_queue_len(&x25_link->async_queue),skb_queue_len(&x25_link->rx_queue));

	PROC_ADD_LINE(m, PROC_LINK_STATS_FORMAT,
		 "Bh Time/Max", atomic_read(&x25_link->bh_time),atomic_read(&x25_link->bh_time_max));
	
	PROC_ADD_LINE(m, link_line);
	PROC_ADD_LINE(m, "| %-15s|  %-38s|\n",
		 "State", x25_link->state == X25_ST_PKT_LVL_RDY_R1 ? "Link Ready" :
			  x25_link->state == X25_ST_RESTART_REQ_SNT_R2 ? "Restart Recvd" :
			  x25_link->state == X25_ST_RESTART_REQ_SNT_R2 ? "Restart Snt" :
			  "Link Not Ready");
	PROC_ADD_LINE(m, link_line);
	PROC_ADD_RET(m);
}

#ifdef LINUX_2_6
static int x25_link_events_get_info(struct seq_file *m, void *v)
#else
static int x25_link_events_get_info(char* buf, char** start, off_t offs, int len)
#endif
{
	x25_link_t *x25_link = PROC_GET_DATA();
	int i;
	unsigned char *event;
	PROC_ADD_DECL(m)

	if (!x25_link)
		return 0;

	PROC_ADD_INIT(m,buf,offs,len);

	for (i=0;i<MAX_PROC_EVENTS;i++){
		event=&x25_link->proc_event_log[i][0];
		if (!strlen(event))
			continue;
		PROC_ADD_LINE(m, "%s \n",event);
	}
	PROC_ADD_RET(m);
}

/*
 *	Prepare data for reading <device> entry.
 *	Return length of data.
 *
 *	On entry, the 'start' argument will contain a pointer to WAN device
 *	data space.
 */


#define PROC_DEV_FORMAT "| %-12s|  %-17d|   %-17d|\n"
#define PROC_DEV_FORMAT1 "| %-15s : %-10d | %-15s : %-10d |\n"
unsigned char dev_line[]= "--------------------------------------------------------\n";
unsigned char dev_line1[]= "---------------------------------------------------------------\n";
unsigned char dev_hdr[] = "| Packet Type |     Received      |    Transmitted     |\n";

#ifdef LINUX_2_6
static int x25_dev_get_info(struct seq_file *m, void *v)
#else
static int x25_dev_get_info(char* buf, char** start, off_t offs, int len)
#endif
{
	x25_lcn_t *x25_lcn = PROC_GET_DATA();
	PROC_ADD_DECL(m)
	
	if (!x25_lcn || !x25_lcn->dev || !x25_lcn->x25_link){
		return 0;
	}

	if (test_bit(LCN_UNREGISTER,&x25_lcn->critical)){
		return 0;
	}
	
	PROC_ADD_INIT(m,buf,offs,len);
	PROC_ADD_LINE(m, dev_line1);
	PROC_ADD_LINE(m, 
		 "|      X25 Protocol Driver Stats                             |\n");
	PROC_ADD_LINE(m, dev_line1);

	PROC_ADD_LINE(m, 
		 "| %-15s : %-10s | %-15s : %-10s |\n", 
		 "Link State", x25_lcn->state == X25_ST_PKT_LVL_RDY_P1 ? "Ready" : 
		 x25_lcn->state ==  X25_ST_CALL_REQ_SNT_P2 ? "Tx Call" :
		 x25_lcn->state ==  X25_ST_CALL_REQ_REC_P3 ? "Rx Call" :
		 x25_lcn->state ==  X25_ST_DATA_TRANSFER_P4 ? "Con.d" :
		 x25_lcn->state ==  X25_ST_CLEAR_REQ_SNT_P6 ? "Tx Clr" :
		 x25_lcn->state ==  X25_ST_CLEAR_REQ_REC_P7 ? "Rx Clr" :
		 x25_lcn->state ==  X25_ST_RESET_REQ_SNT_D2 ? "Tx Rst" :
		 x25_lcn->state ==  X25_ST_RESET_REQ_REC_D3 ? "Rx Rst" :
		 x25_lcn->state ==  X25_ST_PVC_OUT_OF_ORDER ? "PVC Down" :
		 x25_lcn->state ==  X25_ST_HDLC_NOT_IN_ABM ? "Not Ready" : "Invl",
		 "Dev State", netif_queue_stopped(x25_lcn->dev) ? "Stp" : "Run");

	PROC_ADD_LINE(m, PROC_DEV_FORMAT1,
		 "Write Queue", skb_queue_len(&x25_lcn->write_queue),
		 "RR/RNR Queue", skb_queue_len(&x25_lcn->async_queue));

	PROC_ADD_LINE(m, PROC_DEV_FORMAT1, 
		 "Async Queue", skb_queue_len(&x25_lcn->x25_link->async_queue),
		 "Tx BH Kick Cnt", x25_lcn->x25_link->kick_cnt);

	PROC_ADD_LINE(m, 
		 "| %-15s : 0x%-8lx | %-15s : 0x%-8lx |\n", 
		 "X25 TQ Crit", x25_lcn->x25_link->tq_working,
		 "X25 Critical", x25_lcn->critical);
	
	PROC_ADD_LINE(m, 
		 "| %-15s : %-10ld | %-15s : %-10ld |\n", 
		 "BH Tx Running" , (x25_lcn->x25_link->tq_working & BH_RUNNING),	
		 "Dev Tx Aw Kick", (x25_lcn->critical & IF_SEND_AWAITING_KICK));

	PROC_ADD_LINE(m,
		 "| %-15s : %-10ld | %-15s : %-10ld |\n", 
		 "BH Wt Kick" , (x25_lcn->x25_link->tq_working & BH_AWAITING_KICK),	
		 "SVC Link Down", (x25_lcn->x25_link->tq_working & LINK_DOWN));

	PROC_ADD_LINE(m,
		 "| %-15s : %-10d | %-15s : %-10s |\n", 
		 "Bound Lcn" , atomic_read(&x25_lcn->lcn),	
		 "Used By", x25_lcn->x25_sw_lcn ? "Switch" : x25_lcn->sk_id ? "Sock" : x25_lcn->dsp_id ? "DSP" : (x25_lcn->usedby==TCPIP) ? "TCP/IP" : test_bit(LCN_CONFIG_PVC,&x25_lcn->lcn_config) ? "PVC" :test_bit(0,&x25_lcn->used) ? "Busy" : "Free");


	PROC_ADD_LINE(m, PROC_DEV_FORMAT1, 
		 "Lcn RefCnt", atomic_read(&x25_lcn->refcnt),
		 "Dev RefCnt", atomic_read(&x25_lcn->dev->refcnt));

	
	PROC_ADD_LINE(m, dev_line1);
	PROC_ADD_RET(m);
}

/* SNMP
 ******************************************************************************
 *			x25_snmp_data()	
 *
 * Description: Save snmp request and parameters in private structure, enable
 *		TIMER interrupt, put current process in sleep.
 * Arguments:   
 * Returns:
 ******************************************************************************
 */
#define   X25ADMNINDEX          		3
#define   X25ADMNINTERFACEMODE  		4
#define   X25ADMNMAXACTIVECIRCUITS  		5
#define   X25ADMNPACKETSEQUENCING  		6
#define   X25ADMNRESTARTTIMER   		7
#define   X25ADMNCALLTIMER      		8
#define   X25ADMNRESETTIMER     		9
#define   X25ADMNCLEARTIMER     		10
#define   X25ADMNWINDOWTIMER    		11
#define   X25ADMNDATARXMTTIMER  		12
#define   X25ADMNINTERRUPTTIMER  		13
#define   X25ADMNREJECTTIMER    		14
#define   X25ADMNREGISTRATIONREQUESTTIMER  	15
#define   X25ADMNMINIMUMRECALLTIMER  		16
#define   X25ADMNRESTARTCOUNT   		17
#define   X25ADMNRESETCOUNT     		18
#define   X25ADMNCLEARCOUNT     		19
#define   X25ADMNDATARXMTCOUNT  		20
#define   X25ADMNREJECTCOUNT    		21
#define   X25ADMNREGISTRATIONREQUESTCOUNT  	22
#define   X25ADMNNUMBERPVCS     		23
#define   X25ADMNDEFCALLPARAMID  		24
#define   X25ADMNLOCALADDRESS   		25
#define   X25ADMNPROTOCOLVERSIONSUPPORTED  	26
#define   X25OPERINDEX          		29
#define   X25OPERINTERFACEMODE  		30
#define   X25OPERMAXACTIVECIRCUITS  		31
#define   X25OPERPACKETSEQUENCING  		32
#define   X25OPERRESTARTTIMER   		33
#define   X25OPERCALLTIMER      		34
#define   X25OPERRESETTIMER     		35
#define   X25OPERCLEARTIMER     		36
#define   X25OPERWINDOWTIMER    		37
#define   X25OPERDATARXMTTIMER  		38
#define   X25OPERINTERRUPTTIMER  		39
#define   X25OPERREJECTTIMER    		40	
#define   X25OPERREGISTRATIONREQUESTTIMER  	41
#define   X25OPERMINIMUMRECALLTIMER  		42
#define   X25OPERRESTARTCOUNT   		43
#define   X25OPERRESETCOUNT     		44
#define   X25OPERCLEARCOUNT     		45
#define   X25OPERDATARXMTCOUNT  		46
#define   X25OPERREJECTCOUNT    		47
#define   X25OPERREGISTRATIONREQUESTCOUNT  	48
#define   X25OPERNUMBERPVCS     		49
#define   X25OPERDEFCALLPARAMID  		50
#define   X25OPERLOCALADDRESS   		51
#define   X25OPERDATALINKID     		52
#define   X25OPERPROTOCOLVERSIONSUPPORTED  	53
#define   X25STATINDEX          		56
#define   X25STATINCALLS        		57
#define   X25STATINCALLREFUSALS  		58
#define   X25STATINPROVIDERINITIATEDCLEARS  	59
#define   X25STATINREMOTELYINITIATEDRESETS  	60
#define   X25STATINPROVIDERINITIATEDRESETS  	61
#define   X25STATINRESTARTS     		62
#define   X25STATINDATAPACKETS  		63
#define   X25STATINACCUSEDOFPROTOCOLERRORS  	64
#define   X25STATININTERRUPTS   		65
#define   X25STATOUTCALLATTEMPTS  		66
#define   X25STATOUTCALLFAILURES  		67
#define   X25STATOUTINTERRUPTS  		68
#define   X25STATOUTDATAPACKETS  		69
#define   X25STATOUTGOINGCIRCUITS  		70
#define   X25STATINCOMINGCIRCUITS  		71
#define   X25STATTWOWAYCIRCUITS  		72
#define   X25STATRESTARTTIMEOUTS  		73
#define   X25STATCALLTIMEOUTS   		74
#define   X25STATRESETTIMEOUTS  		75
#define   X25STATCLEARTIMEOUTS  		76
#define   X25STATDATARXMTTIMEOUTS  		77
#define   X25STATINTERRUPTTIMEOUTS  		78
#define   X25STATRETRYCOUNTEXCEEDEDS  		79
#define   X25STATCLEARCOUNTEXCEEDEDS  		80
#define   X25CHANNELINDEX       		83
#define   X25CHANNELLIC         		84
#define   X25CHANNELHIC         		85
#define   X25CHANNELLTC         		86
#define   X25CHANNELHTC         		87
#define   X25CHANNELLOC         		88
#define   X25CHANNELHOC         		89
#define   X25CIRCUITINDEX       		92
#define   X25CIRCUITCHANNEL     		93
#define   X25CIRCUITSTATUS      		94
#define   X25CIRCUITESTABLISHTIME  		95
#define   X25CIRCUITDIRECTION   		96
#define   X25CIRCUITINOCTETS    		97
#define   X25CIRCUITINPDUS      		98
#define   X25CIRCUITINREMOTELYINITIATEDRESETS  	99
#define   X25CIRCUITINPROVIDERINITIATEDRESETS  	100
#define   X25CIRCUITININTERRUPTS  		101
#define   X25CIRCUITOUTOCTETS   		102
#define   X25CIRCUITOUTPDUS     		103
#define   X25CIRCUITOUTINTERRUPTS  		104
#define   X25CIRCUITDATARETRANSMISSIONTIMEOUTS  105
#define   X25CIRCUITRESETTIMEOUTS  		106
#define   X25CIRCUITINTERRUPTTIMEOUTS  		107
#define   X25CIRCUITCALLPARAMID  		108
#define   X25CIRCUITCALLEDDTEADDRESS  		109
#define   X25CIRCUITCALLINGDTEADDRESS  		110
#define   X25CIRCUITORIGINALLYCALLEDADDRESS  	111
#define   X25CIRCUITDESCR       		112
#define   X25CLEAREDCIRCUITENTRIESREQUESTED  	113
#define   X25CLEAREDCIRCUITENTRIESGRANTED  	114
#define   X25CLEAREDCIRCUITINDEX  		117
#define   X25CLEAREDCIRCUITPLEINDEX  		118
#define   X25CLEAREDCIRCUITTIMEESTABLISHED  	119
#define   X25CLEAREDCIRCUITTIMECLEARED  	120
#define   X25CLEAREDCIRCUITCHANNEL  		121
#define   X25CLEAREDCIRCUITCLEARINGCAUSE  	122
#define   X25CLEAREDCIRCUITDIAGNOSTICCODE  	123
#define   X25CLEAREDCIRCUITINPDUS  		124
#define   X25CLEAREDCIRCUITOUTPDUS  		125
#define   X25CLEAREDCIRCUITCALLEDADDRESS  	126
#define   X25CLEAREDCIRCUITCALLINGADDRESS  	127
#define   X25CLEAREDCIRCUITCLEARFACILITIES  	128
#define   X25CALLPARMINDEX      		131
#define   X25CALLPARMSTATUS     		132
#define   X25CALLPARMREFCOUNT   		133
#define   X25CALLPARMINPACKETSIZE  		134
#define   X25CALLPARMOUTPACKETSIZE  		135
#define   X25CALLPARMINWINDOWSIZE  		136
#define   X25CALLPARMOUTWINDOWSIZE  		137
#define   X25CALLPARMACCEPTREVERSECHARGING  	138
#define   X25CALLPARMPROPOSEREVERSECHARGING  	139
#define   X25CALLPARMFASTSELECT  		140
#define   X25CALLPARMINTHRUPUTCLASSIZE  	141
#define   X25CALLPARMOUTTHRUPUTCLASSIZE  	142
#define   X25CALLPARMCUG        		143
#define   X25CALLPARMCUGOA      		144
#define   X25CALLPARMBCUG       		145
#define   X25CALLPARMNUI        		146
#define   X25CALLPARMCHARGINGINFO  		147
#define   X25CALLPARMRPOA       		148
#define   X25CALLPARMTRNSTDLY   		149
#define   X25CALLPARMCALLINGEXT  		150
#define   X25CALLPARMCALLEDEXT  		151
#define   X25CALLPARMINMINTHUPUTCLS  		152
#define   X25CALLPARMOUTMINTHUPUTCLS  		153
#define   X25CALLPARMENDTRNSDLY  		154
#define   X25CALLPARMPRIORITY   		155
#define   X25CALLPARMPROTECTION  		156
#define   X25CALLPARMEXPTDATA   		157
#define   X25CALLPARMUSERDATA   		158
#define   X25CALLPARMCALLINGNETWORKFACILITIES  	159
#define   X25CALLPARMCALLEDNETWORKFACILITIES  	160

int x25_snmp_data(x25_lcn_t* x25_lcn, struct ifreq* ifr)
{
	wanpipe_snmp_t	snmp;
	x25_link_t*	x25_link = NULL;
	
	printk(KERN_INFO "DEBUG: x25_snmp_data\n");
	x25_link = x25_lcn->x25_link;
	
	if(copy_from_user(&snmp, ifr->ifr_data, sizeof(wanpipe_snmp_t))) {
		return -EFAULT;
	}

	switch(snmp.snmp_magic){
	/********** X.25 Administration Table *********/
	/*case X25ADMNINDEX:*/
	case X25ADMNINTERFACEMODE:
		snmp.snmp_val = 
			(x25_link->cfg.dte) ? SNMP_X25_DTE : SNMP_X25_DCE;
		break;

	case X25ADMNMAXACTIVECIRCUITS:
	        snmp.snmp_val = SNMP_X25_MAX_CIRCUIT_NUM;
		break;

	case X25ADMNPACKETSEQUENCING:
        	snmp.snmp_val = (x25_link->cfg.mode = GFI_FOR_MOD_8) ? SNMP_X25_MODULO8 : SNMP_X25_MODULO128;
		break;

	case X25ADMNRESTARTTIMER:
		snmp.snmp_val = x25_link->cfg.T10_T20;
		break;

	case X25ADMNCALLTIMER:
		printk(KERN_INFO "DEBUG: CALLTIMER\n");
		snmp.snmp_val = x25_link->cfg.T11_T21;
		break;

	case X25ADMNRESETTIMER:
		snmp.snmp_val = x25_link->cfg.T12_T22;
		break;

	case X25ADMNCLEARTIMER:
		snmp.snmp_val = x25_link->cfg.T13_T23;
		break;

	case X25ADMNWINDOWTIMER: /* FIXME */
		snmp.snmp_val = 0;
		break;

	case X25ADMNDATARXMTTIMER: /* FIXME */
		snmp.snmp_val = 0;
		break;

	case X25ADMNINTERRUPTTIMER:
		snmp.snmp_val = x25_link->cfg.T16_T26;
		break;

	case X25ADMNREJECTTIMER: /* FIXME */
		snmp.snmp_val = 0;
		break;

	case X25ADMNREGISTRATIONREQUESTTIMER:
		snmp.snmp_val = x25_link->cfg.T28;
		break;

	case X25ADMNMINIMUMRECALLTIMER: /* FIXME */
		snmp.snmp_val = 0;
		break;

	case X25ADMNRESTARTCOUNT:
		snmp.snmp_val = x25_link->cfg.R10_R20;
		break;

	case X25ADMNRESETCOUNT:
		snmp.snmp_val = x25_link->cfg.R12_R22;
		break;

	case X25ADMNCLEARCOUNT:    	
		snmp.snmp_val = x25_link->cfg.R13_R23;
		break;

	case X25ADMNDATARXMTCOUNT: /* FIXME */
		snmp.snmp_val = 0;
		break;

	case X25ADMNREJECTCOUNT: /* FIXME */
		snmp.snmp_val = 0;
		break;

	case X25ADMNREGISTRATIONREQUESTCOUNT: /* FIXME */
		snmp.snmp_val = 0;
		break;

	case X25ADMNNUMBERPVCS:  /* FIXME */   		
		snmp.snmp_val = 0;
		break;

	case X25ADMNDEFCALLPARAMID: /* FIXME */
		snmp.snmp_val = 0;
		break;

	case X25ADMNLOCALADDRESS: /* FIXME */
		snmp.snmp_val = 0;
		break;

	case X25ADMNPROTOCOLVERSIONSUPPORTED: /* FIXME */
		snmp.snmp_val = 0;
		break;

	/********** X.25 Operational Table *********/
	/*case X25OPERINDEX:*/
	case X25OPERINTERFACEMODE:	
		snmp.snmp_val = (x25_link->cfg.dte) ? SNMP_X25_DTE : SNMP_X25_DCE;
		break;

	case X25OPERMAXACTIVECIRCUITS: /* FIXME */ 	
		snmp.snmp_val = SNMP_X25_MAX_CIRCUIT_NUM;
		break;

	case X25OPERPACKETSEQUENCING:  
		snmp.snmp_val = (x25_link->cfg.mode = GFI_FOR_MOD_8) ? SNMP_X25_MODULO8 : SNMP_X25_MODULO128;
		break;

	case X25OPERRESTARTTIMER:   	
		snmp.snmp_val = x25_link->cfg.T10_T20;
		break;

	case X25OPERCALLTIMER:      	
		snmp.snmp_val = x25_link->cfg.T11_T21;
		break;

	case X25OPERRESETTIMER:     
		snmp.snmp_val = x25_link->cfg.T12_T22;
		break;

	case X25OPERCLEARTIMER:    
		snmp.snmp_val = x25_link->cfg.T13_T23;
		break;

	case X25OPERWINDOWTIMER: /* FIXME */
		snmp.snmp_val = 0;
		break;

	case X25OPERDATARXMTTIMER: /* FIXME */
		snmp.snmp_val = 0;
		break;

	case X25OPERINTERRUPTTIMER: 
		snmp.snmp_val = x25_link->cfg.T16_T26;
		break;

	case X25OPERREJECTTIMER: /* FIXME */
		snmp.snmp_val = 0;
		break;

	case X25OPERREGISTRATIONREQUESTTIMER: 
                snmp.snmp_val = x25_link->cfg.T28;
		break;

	case X25OPERMINIMUMRECALLTIMER: /* FIXME */
                snmp.snmp_val = 0;
		break;

	case X25OPERRESTARTCOUNT:  
                snmp.snmp_val = x25_link->cfg.R10_R20;
		break;

	case X25OPERRESETCOUNT:    
                snmp.snmp_val = x25_link->cfg.R12_R22;
		break;

	case X25OPERCLEARCOUNT:    
                snmp.snmp_val = x25_link->cfg.R13_R23;
		break;

	case X25OPERDATARXMTCOUNT: /* FIXME */
                snmp.snmp_val = 0;
		break;

	case X25OPERREJECTCOUNT: /* FIXME */
                snmp.snmp_val = 0;
		break;

	case X25OPERREGISTRATIONREQUESTCOUNT: /* FIXME */
                snmp.snmp_val = 0;
		break;

	case X25OPERNUMBERPVCS: /* FIXME */    
                snmp.snmp_val = x25_link->cfg.highest_PVC;
		break;

	case X25OPERDEFCALLPARAMID: /* FIXME */
                snmp.snmp_val = 0;
		break;
	
	case X25OPERLOCALADDRESS: /* FIXME */
                strcpy((void*)snmp.snmp_data, "N/A");
		break;

	case X25OPERDATALINKID: /* FIXME */
                strcpy((void*)snmp.snmp_data, "N/A");
		break;

	case X25OPERPROTOCOLVERSIONSUPPORTED: /* FIXME */
                strcpy((void*)snmp.snmp_data, "N/A");
		break;


		 
	/********** X.25 Statistics Table *********/
	/*case X25STATINDEX:*/         
	case X25STATINCALLS:       
        	snmp.snmp_val = x25_link->x25stats.rx_call_req;
		break;

	case X25STATINCALLREFUSALS: 
                snmp.snmp_val = x25_link->x25stats.tx_clear_req;
		break;

	case X25STATINPROVIDERINITIATEDCLEARS: 
                snmp.snmp_val = (x25_link->cfg.dte) ? 
				x25_link->x25stats.rx_clear_req : 
				x25_link->x25stats.tx_clear_req;
		break;

	case X25STATINREMOTELYINITIATEDRESETS:  
                snmp.snmp_val = (x25_link->cfg.dte) ? 
				x25_link->x25stats.tx_reset_req : 
				x25_link->x25stats.rx_reset_req;
		break;

	case X25STATINPROVIDERINITIATEDRESETS:  
                snmp.snmp_val = (x25_link->cfg.dte) ? 
				x25_link->x25stats.rx_reset_req : 
				x25_link->x25stats.tx_reset_req;
		break;

	case X25STATINRESTARTS:     
                snmp.snmp_val = x25_link->x25stats.rx_restrt_req;
		break;

	case X25STATINDATAPACKETS:  
                snmp.snmp_val = x25_link->x25stats.rx_data_pkt;
		break;

	case X25STATINACCUSEDOFPROTOCOLERRORS:  
                snmp.snmp_val = x25_link->x25stats.rx_clear_req + 
				x25_link->x25stats.rx_reset_req +
				x25_link->x25stats.rx_restrt_req +
				x25_link->x25stats.rx_diag;
		break;

	case X25STATININTERRUPTS:   
                snmp.snmp_val = x25_link->x25stats.rx_intr_req;
		break;

	case X25STATOUTCALLATTEMPTS:  
                snmp.snmp_val = x25_link->x25stats.tx_call_req;
		break;

	case X25STATOUTCALLFAILURES:  
                snmp.snmp_val = x25_link->x25stats.tx_clear_req;
		break;

	case X25STATOUTINTERRUPTS: 
                snmp.snmp_val = x25_link->x25stats.tx_intr_req;
		break;

	case X25STATOUTDATAPACKETS: 
                snmp.snmp_val = x25_link->x25stats.tx_data_pkt;
		break;

	case X25STATOUTGOINGCIRCUITS: 
                snmp.snmp_val = x25_link->x25stats.tx_call_req-
				x25_link->x25stats.rx_clear_req;
		break;

	case X25STATINCOMINGCIRCUITS:  
                snmp.snmp_val = x25_link->x25stats.rx_call_req-
				x25_link->x25stats.tx_clear_req;
		break;

	case X25STATTWOWAYCIRCUITS:  
                snmp.snmp_val = x25_link->x25stats.tx_call_req-
				x25_link->x25stats.rx_clear_req;
		break;

	case X25STATRESTARTTIMEOUTS: /* FIXME */
                snmp.snmp_val = 0;
		break;

	case X25STATCALLTIMEOUTS: /* FIXME */
                snmp.snmp_val = 0;
		break;

	case X25STATRESETTIMEOUTS: /* FIXME */
                snmp.snmp_val = 0;
		break;

	case X25STATCLEARTIMEOUTS: /* FIXME */
                snmp.snmp_val = 0;
		break;

	case X25STATDATARXMTTIMEOUTS: /* FIXME */
                snmp.snmp_val = 0;
		break;

	case X25STATINTERRUPTTIMEOUTS: /* FIXME */
                snmp.snmp_val = 0;
		break;

	case X25STATRETRYCOUNTEXCEEDEDS: /* FIXME */
                snmp.snmp_val = 0;
		break;

	case X25STATCLEARCOUNTEXCEEDEDS: /* FIXME */
                snmp.snmp_val = 0;
		break;

	/********** X.25 Channel Table *********/
	/* case X25CHANNELINDEX: */
	case X25CHANNELLIC:        
		snmp.snmp_val = x25_link->cfg.lowest_incoming_channel;
		break;

	case X25CHANNELHIC:         
                snmp.snmp_val = x25_link->cfg.highest_incoming_channel;
		break;

	case X25CHANNELLTC:        
                snmp.snmp_val = x25_link->cfg.lowest_two_way_channel;
		break;

	case X25CHANNELHTC:        
                snmp.snmp_val = x25_link->cfg.highest_two_way_channel;
		break;

	case X25CHANNELLOC:        
                snmp.snmp_val = x25_link->cfg.lowest_outgoing_channel;
		break;

	case X25CHANNELHOC:         
                snmp.snmp_val = x25_link->cfg.highest_outgoing_channel;
		break;

	/********** X.25 Per Circuits Information Table *********/
	/*case X25CIRCUITINDEX:*/
	case X25CIRCUITCHANNEL: /* FIXME */
               	snmp.snmp_val = 0;
		break;

	case X25CIRCUITSTATUS: /* FIXME */ 
                snmp.snmp_val = 0;
		break;

	case X25CIRCUITESTABLISHTIME: /* FIXME */
                snmp.snmp_val = 0;	//chan->chan_establ_time;
		break;

	case X25CIRCUITDIRECTION:   /* FIXME */
                snmp.snmp_val = 0;	//chan->chan_direct;
		break;

	case X25CIRCUITINOCTETS: 
                snmp.snmp_val = x25_lcn->ifstats.rx_bytes;
		break;

	case X25CIRCUITINPDUS:      
                snmp.snmp_val = x25_lcn->ifstats.rx_packets;
		break;

	case X25CIRCUITINREMOTELYINITIATEDRESETS: /* FIXME */
                snmp.snmp_val = 0;
		break;

	case X25CIRCUITINPROVIDERINITIATEDRESETS: /* FIXME */        
                snmp.snmp_val = 0;
		break;

	case X25CIRCUITININTERRUPTS:  
                snmp.snmp_val = x25_link->x25stats.rx_intr_req;
		break;

	case X25CIRCUITOUTOCTETS:   
                snmp.snmp_val = x25_lcn->ifstats.tx_bytes;
		break;

	case X25CIRCUITOUTPDUS:     
                snmp.snmp_val = x25_lcn->ifstats.tx_packets;
		break;

	case X25CIRCUITOUTINTERRUPTS:  
                snmp.snmp_val = x25_link->x25stats.tx_intr_req;
		break;

	case X25CIRCUITDATARETRANSMISSIONTIMEOUTS: /* FIXME */	       
                snmp.snmp_val = 0;
		break;

	case X25CIRCUITRESETTIMEOUTS: /* FIXME */	       
                snmp.snmp_val = 0;
		break;

	case X25CIRCUITINTERRUPTTIMEOUTS: /* FIXME */	       
                snmp.snmp_val = 0;
		break;

	case X25CIRCUITCALLPARAMID: /* FIXME */	       
                snmp.snmp_val = 0;
		break;

	case X25CIRCUITCALLEDDTEADDRESS: /* FIXME */       
                strcpy((void*)snmp.snmp_data, "N/A");
		break;

	case X25CIRCUITCALLINGDTEADDRESS: /* FIXME */	       
                strcpy((void*)snmp.snmp_data, "N/A");
		break;

	case X25CIRCUITORIGINALLYCALLEDADDRESS: /* FIXME */       
                snmp.snmp_val = 0;
		break;

	case X25CIRCUITDESCR: /* FIXME */      
                strcpy((void*)snmp.snmp_data, "N/A");
		break;


	case X25CLEAREDCIRCUITENTRIESREQUESTED: /* FIXME */
                snmp.snmp_val = 0;
		break;

	case X25CLEAREDCIRCUITENTRIESGRANTED: /* FIXME */
                snmp.snmp_val = 0;
		break;

	/********** X.25 The Cleared Circuit Table *********/
	case X25CLEAREDCIRCUITINDEX: /* FIXME */
        	snmp.snmp_val = 0;
		break;

	case X25CLEAREDCIRCUITPLEINDEX: /* FIXME */
                snmp.snmp_val = 0;
		break;

	case X25CLEAREDCIRCUITTIMEESTABLISHED: /* FIXME */
                snmp.snmp_val = 0;	//chan->chan_establ_time;
		break;

	case X25CLEAREDCIRCUITTIMECLEARED: /* FIXME */
                snmp.snmp_val = 0;	//chan->chan_clear_time;
		break;

	case X25CLEAREDCIRCUITCHANNEL:  /* FIXME */
                snmp.snmp_val = 0;	//chan->ch_idx;
		break;

	case X25CLEAREDCIRCUITCLEARINGCAUSE: /* FIXME */ 
                snmp.snmp_val = 0;	//chan->chan_clear_cause;
		break;

	case X25CLEAREDCIRCUITDIAGNOSTICCODE: /* FIXME */
                snmp.snmp_val = 0;	//chan->chan_clear_diagn;
		break;

	case X25CLEAREDCIRCUITINPDUS: 
                snmp.snmp_val = x25_link->x25stats.rx_data_pkt;
		break;

	case X25CLEAREDCIRCUITOUTPDUS:  
                snmp.snmp_val = x25_link->x25stats.tx_data_pkt;
		break;

	case X25CLEAREDCIRCUITCALLEDADDRESS: /* FIXME */ 
                strcpy((void*)snmp.snmp_data, "N/A"); 
		break;

	case X25CLEAREDCIRCUITCALLINGADDRESS:  /* FIXME */
                strcpy((void*)snmp.snmp_data, "N/A"); 
		break;

	case X25CLEAREDCIRCUITCLEARFACILITIES:  
                strcpy((void*)snmp.snmp_data, "N/A");
		break;

	/********** X.25 The Call Parameter Table *********/
	case X25CALLPARMINDEX: /* FIXME */
                snmp.snmp_val = 0;
		break;

	case X25CALLPARMSTATUS: /* FIXME */
                snmp.snmp_val = 0;
		break;

	case X25CALLPARMREFCOUNT: /* FIXME */
                snmp.snmp_val = 0;
		break;

	case X25CALLPARMINPACKETSIZE:  
                snmp.snmp_val = x25_lcn->rx_pkt_size;
		break;

	case X25CALLPARMOUTPACKETSIZE: 
                snmp.snmp_val = x25_lcn->tx_pkt_size;
		break;

	case X25CALLPARMINWINDOWSIZE:
                snmp.snmp_val = x25_lcn->rx_window_size;
		break;

	case X25CALLPARMOUTWINDOWSIZE: 
                snmp.snmp_val = x25_lcn->tx_window_size;
		break;

	case X25CALLPARMACCEPTREVERSECHARGING: /* FIXME */
                snmp.snmp_val = SNMP_X25_ARC_DEFAULT;
		break;

	case X25CALLPARMPROPOSEREVERSECHARGING: /* FIXME */
                snmp.snmp_val = SNMP_X25_PRC_DEFAULT;
		break;

	case X25CALLPARMFASTSELECT: /* FIXME */
                snmp.snmp_val = SNMP_X25_FS_DEFAULT;
		break;

	case X25CALLPARMINTHRUPUTCLASSIZE: /* FIXME */
                snmp.snmp_val = SNMP_X25_THRUCLASS_TCDEF;
		break;

	case X25CALLPARMOUTTHRUPUTCLASSIZE: /* FIXME */
                snmp.snmp_val = SNMP_X25_THRUCLASS_TCDEF;
		break;

	case X25CALLPARMCUG: /* FIXME */
                strcpy((void*)snmp.snmp_data, "N/A");
		break;

	case X25CALLPARMCUGOA: /* FIXME */
                strcpy((void*)snmp.snmp_data, "N/A");
		break;

	case X25CALLPARMBCUG: /* FIXME */
                strcpy((void*)snmp.snmp_data, "N/A");
		break;

	case X25CALLPARMNUI: /* FIXME */
                strcpy((void*)snmp.snmp_data, "N/A");
		break;

	case X25CALLPARMCHARGINGINFO: /* FIXME */
                snmp.snmp_val = SNMP_X25_CHARGINGINFO_DEF;
		break;

	case X25CALLPARMRPOA: /* FIXME */
                strcpy((void*)snmp.snmp_data, "N/A");
		break;

	case X25CALLPARMTRNSTDLY: /* FIXME */
                snmp.snmp_val = 0;
		break;

	case X25CALLPARMCALLINGEXT: /* FIXME */
                strcpy((void*)snmp.snmp_data, "N/A");
		break;

	case X25CALLPARMCALLEDEXT: /* FIXME */
                strcpy((void*)snmp.snmp_data, "N/A");
		break;

	case X25CALLPARMINMINTHUPUTCLS: /* FIXME */
                snmp.snmp_val = SNMP_X25_THRUCLASS_TCDEF;
		break;

	case X25CALLPARMOUTMINTHUPUTCLS: /* FIXME */
                snmp.snmp_val = SNMP_X25_THRUCLASS_TCDEF;
		break;

	case X25CALLPARMENDTRNSDLY: /* FIXME */
                strcpy((void*)snmp.snmp_data, "N/A");
		break;

	case X25CALLPARMPRIORITY: /* FIXME */
                strcpy((void*)snmp.snmp_data, "N/A");
		break;

	case X25CALLPARMPROTECTION: /* FIXME */
                strcpy((void*)snmp.snmp_data, "N/A");
		break;

	case X25CALLPARMEXPTDATA: /* FIXME */
                snmp.snmp_val = SNMP_X25_EXPTDATA_DEFULT;
		break;

	case X25CALLPARMUSERDATA:   
                strcpy((void*)snmp.snmp_data, x25_lcn->accept_udata);
		break;

	case X25CALLPARMCALLINGNETWORKFACILITIES: 
                strcpy((void*)snmp.snmp_data, x25_lcn->accept_calling);
		break;

	case X25CALLPARMCALLEDNETWORKFACILITIES: 
                strcpy((void*)snmp.snmp_data, x25_lcn->accept_called);
		break;

	default:
            	return -EAFNOSUPPORT;
	}

	if(copy_to_user(ifr->ifr_data, &snmp, sizeof(wanpipe_snmp_t)))
		return -EFAULT;

	return 0;

}


#endif


