#include <linux/wanpipe_includes.h>
#include <linux/wanpipe_defines.h>
#include <linux/wanpipe_debug.h>
#include <linux/wanpipe_common.h>
#include <linux/wanpipe_cfg.h>
#include "wanpipe_x25.h"

x25_lcn_t* x25_get_switch_device(x25_link_t* x25_link, x25_lcn_t* x25_lcn, struct sk_buff* skb);
int x25_sw_replace_addr(x25_lcn_t* x25_lcn, struct sk_buff* skb, int flag);

inline x25_lcn_t* x25_get_free_sw_dev (x25_link_t *x25_link, x25_lcn_t *x25_orig_lcn, unsigned short lcn)
{
	struct net_device *dev;
	x25_lcn_t *x25_lcn=NULL;
	unsigned long flags;

	WP_READ_LOCK(&x25_link->dev_list_lock,flags);
	
	for (dev=x25_link->dev_list; dev; dev = *(struct net_device **)dev->priv){
		x25_lcn = (x25_lcn_t *)dev->priv;

		if (!x25_lcn || !(dev->flags&IFF_UP) || (x25_orig_lcn == x25_lcn))
			continue;

		if (x25_lcn->usedby == API && 
		    !x25_lcn->dsp_id &&
		    !x25_lcn->sk_id &&
		    !x25_lcn->x25_sw_lcn &&
		    x25_lcn->state == X25_ST_PKT_LVL_RDY_P1 &&
		    atomic_read(&x25_lcn->lcn) == 0 &&
		    atomic_read(&x25_lcn->refcnt) == 1 &&
		    !test_and_set_bit(0,&x25_lcn->used)){

			atomic_set(&x25_lcn->lcn,lcn);

			//FIXME LOCK
			x25_lcn->x25_sw_lcn	 = (void*)x25_orig_lcn;	
			X25_HOLD(x25_orig_lcn);
			
			x25_orig_lcn->x25_sw_lcn = (void*)x25_lcn;
			X25_HOLD(x25_lcn);

			break;
		}
	}

	WP_READ_UNLOCK(&x25_link->dev_list_lock,flags);


	if (!dev){
		return NULL;
	}

	return x25_lcn;
}

int x25_sw_call_request(x25_link_t* x25_link, struct net_device* dev, struct sk_buff* skb, struct x25_frame* frame)
{
	x25_lcn_t*		x25_lcn = (x25_lcn_t*)dev->priv;
	x25_lcn_t*		x25_sw_lcn = NULL;

	if (!x25_lcn){
		return X25_INVALUE;
	}

	
	if (x25_lcn->x25_sw_lcn != NULL){
		printk(KERN_INFO "%s: Logical channel %s: already switching..\n",
				x25_lcn->dev->name,((x25_lcn_t*)x25_lcn->x25_sw_lcn)->dev->name);
		return X25_INVALUE;
	}

	/* Search for the switch LCN */
	if ((x25_sw_lcn = x25_get_switch_device(x25_link, x25_lcn, skb)) == NULL){
		return X25_INVALUE;
	}

	/* Replace real address by fake addresss */
	x25_sw_replace_addr(x25_sw_lcn, skb, X25_SW_FWD);

	x25_sw_lcn->x25_api_cmd_hdr.pktType	= frame->type;
	x25_sw_lcn->x25_api_cmd_hdr.lcn		= frame->lcn;
	x25_sw_lcn->x25_api_cmd_hdr.length	= strlen(x25_sw_lcn->call_string);

	if (x25_place_call(x25_sw_lcn) != 0){
		x25_release_lcn_dev(x25_sw_lcn->x25_link, x25_sw_lcn);
		return X25_INVALUE;
	}

	/* Connect to svc's */
	return X25_OK;
}

int x25_sw_accept_call(x25_lcn_t* x25_lcn, struct sk_buff* skb, struct x25_frame* frame)
{
	x25_lcn_t*	x25_sw_lcn = NULL;


	if (x25_lcn->x25_sw_lcn == NULL){
		printk(KERN_INFO "Debug: No switching device\n");
		return X25_INVALUE;
	}
	x25_sw_lcn = x25_lcn->x25_sw_lcn;

	/* Replace real address by fake addresss */
	x25_sw_replace_addr(x25_sw_lcn, skb, X25_SW_BACK);

	x25_sw_lcn->x25_api_cmd_hdr.pktType	= frame->type;
	x25_sw_lcn->x25_api_cmd_hdr.lcn		= frame->lcn;
	x25_sw_lcn->x25_api_cmd_hdr.length	= strlen(x25_sw_lcn->call_string);

	if (x25_accept_call(x25_sw_lcn)){
		return X25_INVALUE;
	}
	
	return X25_OK;
}

int x25_sw_restart_request(x25_link_t* x25_link)
{
	struct net_device*	dev = NULL;
	x25_lcn_t*		x25_lcn = NULL;
	struct x25_frame	frame;
	unsigned long flags;

	frame.type = CLEAR_REQUEST_PKT;
	frame.cause = 0x00;
	frame.diag = 0x00;

	WP_READ_LOCK(&x25_link->dev_list_lock,flags);
	
	for (dev=x25_link->dev_list; dev; dev = *(struct net_device **)dev->priv){
		x25_lcn = (x25_lcn_t *)dev->priv;
		if (x25_lcn && x25_lcn->x25_sw_lcn){
			x25_sw_send_control(x25_lcn->x25_sw_lcn, NULL, &frame);
			/* Do not clear the x25_sw_lcn pointers here, they
			 * will be cleared on clear confirmation */
		}
	}

	WP_READ_UNLOCK(&x25_link->dev_list_lock,flags);

	return X25_OK;
}

int x25_sw_data_rx(x25_lcn_t* x25_lcn, struct sk_buff* skb, struct x25_frame* frame)
{
	x25api_hdr_t *x25_api_hdr=NULL;
	x25_lcn_t* x25_sw_lcn = (x25_lcn_t*)x25_lcn->x25_sw_lcn;
	int len=skb->len;

	if (x25_sw_lcn == NULL){
		printk(KERN_INFO "%s: Error: Switching device is NULL\n",
				x25_lcn->dev->name);
		return -EINVAL;
	}

	if (test_bit(LCN_DOWN,&x25_sw_lcn->critical)){
		printk(KERN_INFO "%s: Error Switching device is DOWN\n",
				x25_sw_lcn->dev->name);
		return -EINVAL;
	}
	
	if (skb_headroom(skb) < sizeof(x25api_hdr_t)){
		printk(KERN_INFO "x25_sw_data_tx_up: NO Room for X25 Header!\n");
		return -EINVAL;
	}
	
	x25_api_hdr = (x25api_hdr_t*)skb_push(skb,sizeof(x25api_hdr_t));

	x25_api_hdr->qdm = (frame->gfi >> 1);
	if (frame->mbit){
		set_bit(0,(unsigned long*)&x25_api_hdr->qdm);
	}else{
		clear_bit(0,(unsigned long*)&x25_api_hdr->qdm);
	}

	x25_api_hdr->cause = frame->cause;
	x25_api_hdr->diagn = frame->diag;
	x25_api_hdr->pktType = frame->type;
	x25_api_hdr->length  = 0;
	x25_api_hdr->result = 0;
	x25_api_hdr->lcn = atomic_read(&x25_sw_lcn->lcn);
	
	if (netif_queue_stopped(x25_sw_lcn->dev)){
		printk(KERN_INFO "%s: Ooops.....  should never happened!\n",
					x25_sw_lcn->dev->name);
		return -EBUSY;
	}

	X25_SKB_DEC(skb->truesize);
	if (x25_sw_lcn->dev->hard_start_xmit(skb,x25_sw_lcn->dev) == X25_OK){
	    	//printk(KERN_INFO "Debug: Data sent to switching device\n");
	}else{
	    	printk(KERN_INFO "Debug: %s: Failed to sent data to switching device\n",
				x25_sw_lcn->dev->name);
		X25_SKB_INC(skb->truesize);
		return -EBUSY;
	}

	x25_lcn->ifstats.rx_packets++;
	x25_lcn->ifstats.rx_bytes += len;
	return 0;
		
}

int x25_sw_send_control(x25_lcn_t* x25_sw_lcn, struct sk_buff* skb, struct x25_frame* frame)
{
	/* Update switching Pr count only for RR packet */
	if (frame->type == RR_PKT)
		x25_sw_lcn->sw_Pr_count = frame->Pr_count;

	x25_send_control(x25_sw_lcn->x25_link,0,
			 atomic_read(&x25_sw_lcn->lcn), frame->type, 
			 frame->cause, frame->diag, 
			 skb ? skb->data : NULL, skb ? skb->len : 0);
	return X25_OK;
}

x25_lcn_t* x25_get_switch_device(x25_link_t *x25_link, x25_lcn_t* x25_lcn, struct sk_buff* skb)
{
	x25_link_t*		x25_sw_link = NULL;
	unsigned char* 		buf = NULL;
	unsigned char		addr[WAN_ADDRESS_SZ+1];
	int 			len = 0;

	buf = strstr(skb->data, "-d");
	if (buf == NULL)
		return NULL;

	/* Skip "-d" from the string */
	buf += 2;
	memset(addr, '\0', WAN_ADDRESS_SZ+1);
	
	/* Get destination address length */
	while(buf[len] != ' ' && buf[len] != '\0') len++;
	memcpy(addr, buf, len);
	
	/* Search for "fake" address through all X.25 links */
	if ((x25_sw_link = x25_get_switch_link(addr)) == NULL){
		printk(KERN_INFO "%s: Switch Fake address (%s) not found: Clearing Call!\n",
				x25_link->name,addr);
		return NULL;
	}

	if (x25_sw_link == x25_link){
		printk(KERN_INFO "Error: Switch Link=%s  == Orig Link=%s\n",
				x25_sw_link->name, x25_link->name);
		return NULL;
	}

	return x25_get_free_sw_dev(x25_sw_link,x25_lcn,0xFFFF);
}

int x25_sw_replace_addr(x25_lcn_t* x25_lcn, struct sk_buff* skb, int flag)
{
	x25_link_t* 	x25_link = x25_lcn->x25_link;
	unsigned char* 	buf = NULL;
	int		len = 0;

	memset(x25_lcn->call_string, '\0', X25_CALL_STR_SZ+1);
	if ((buf = strstr(skb->data, "-d"))){
		strncat(x25_lcn->call_string, "-d", 2);
		if (flag == X25_SW_FWD)
			strcat(x25_lcn->call_string, x25_link->real_addr);
		else
			strcat(x25_lcn->call_string, x25_link->virtual_addr);
	}
	if ((buf = strstr(skb->data, "-s"))){
		len = 0;
		while(buf[len] != ' ' && buf[len] != '\0') len++;
		strncat(x25_lcn->call_string, " ", 1);
		strncat(x25_lcn->call_string, buf, len);
	}
	if ((buf = strstr(skb->data, "-f"))){
		len = 0;
		while(buf[len] != ' ' && buf[len] != '\0') len++;
		strncat(x25_lcn->call_string, " ", 1);
		strncat(x25_lcn->call_string, buf, len);
	}
	if ((buf = strstr(skb->data, "-u"))){
		len = 0;
		while(buf[len] != ' ' && buf[len] != '\0') len++;
		strncat(x25_lcn->call_string, " ", 1);
		strncat(x25_lcn->call_string, buf, len);
	}
	return 0;
}

void x25_sw_disconnect(x25_lcn_t *x25_sw_lcn)
{
	
	if (!x25_sw_lcn)
		return;

	if (atomic_read(&x25_sw_lcn->lcn)==0 || atomic_read(&x25_sw_lcn->lcn)==0xFFFF)
		return;

	switch (x25_sw_lcn->state){
	
		case X25_ST_PKT_LVL_RDY_P1:
		case X25_ST_CLEAR_REQ_SNT_P6:
		case X25_ST_HDLC_NOT_IN_ABM:
			return;

		default:
			break;
	}

	lcn_state_change(x25_sw_lcn,X25_ST_CLEAR_REQ_SNT_P6);

	x25_send_control(x25_sw_lcn->x25_link,
			 0,
			 atomic_read(&x25_sw_lcn->lcn),
			 CLEAR_REQUEST_PKT, 
			 0,OSI_NETWORK_PROBLEM_PERMANENT_DISC,
			 NULL,0);
}

